#!/usr/bin/env python2estimd
# -*- coding: utf-8 -*-
"""
Created on Sat Dec  9 08:00:56 2017

@author: Tomas McKelvey
"""

import numpy as np
from scipy import linalg

def fsid():
    """
    This is the FSID toolbox for Matlab version 1.3, 2021-07-14

    Functions: fdsid, gfdsid, ffsid, gffsid, lrm

    FSID is open source software licenced under the GNU GPL v.3
    FSID can be accessed at https://github.com/tomasmckelvey/fsid

    """
    print("FSID for python version 1.3, 2021-07-14")


def lrm(u, y, n=2 , Nw=None, DFTindices=None, lam=0):
    """
        Calulates the non-parametric frequency MIMO function estimate using the Local Rational Model mathod.

    Parameters
    ----------
    u : array
        input samples where element u[i,j] holds the sample for time index i and input channel j
    y : array
        output samples where element y[i,j] holds the sample for time index i and input channel j
    n : integer
        local rational model order (default n=2)
    Nw :integer (optional)
        determines local frequency window. To determine the estimate at DFT frequency index k,
        DFT samples from index k-Nw to k+Nw are employed (with wraparound at the edges)
        if Nw = None (default) then Nw Nw = 2*np.int(np.ceil(((m+2)*n+1.0)/2)) where m is the number of input channels
    DFTindices : array of integers
        a subset of the DFT indices. The local model is inlu calculated for frequency indices in this set
    lam : real valued scalar. (optinal)
        if lam>0 then the LS-problem is augmented with parameter penalty scaled with lam*norm(R)
        where R is the associated regression matrix.

    Returns
    -------
    ff : array of frequency function estimate where element ff[k,i,j] correspond to DFT frequency index k,
        output index i and input index j. If DFTindices is given then ff[i,i,j] holds the frequency function
        estimate at DFT frequency index DFTindices[i]

    """
    us = np.shape(u)
    if len(us) == 1:
        m = 1
        u = np.reshape(u,(us[0],1))
    else:
        m = us[1]
    Nu = us[0]
    ys = np.shape(y)
    if len(ys) == 1:
        p = 1
        y = np.reshape(y,(ys[0],1))
    else:
        p = ys[1]
    Ny = ys[0]
    if Nu != Ny:
        print("lrm: incompatible number of time samples in u and y")
        return False
    if Nw is None:
        Nw = 2*int(np.ceil(((m+2)*n+1.0)/2))
    #print("Nw ",Nw)
    if Nw <int(np.ceil(((m+2)*n+1.0)/2)):
        print("Error: Nw too small.")
        return False
    yf = np.fft.fft(y, axis=0)
    uf = np.fft.fft(u, axis=0)
    R = np.vander(np.arange(-Nw,Nw+1),n+1)
#    print("NW", Nw)
#    print("size R", np.shape(R))
    yfe = np.vstack((yf[-Nw:,:], yf, yf[:Nw] ))
    ufe = np.vstack((uf[-Nw:,:], uf, uf[:Nw] ))
    iset = np.arange(-Nw,Nw+1)
    if DFTindices is None:
        DFTindices = np.arange(Ny)
#    for i in np.arange(Ny):
    ff = np.zeros((np.size(DFTindices), p, m), dtype=complex)
    for fi, i in enumerate(DFTindices):
        for pidx in range(p):
            # A(z) yf = B(z) uf + T(z)
            yy = yfe[Nw+i+iset,pidx]
 #           uu = ufe[Nw+i+iset,0]
            RR = np.hstack((R, np.matmul(np.diag(-yy), R[:,:n])  ))
            for midx in np.flip(range(m),0):
                RR = np.hstack((RR, np.matmul(np.diag(ufe[Nw+i+iset, midx]), R)))
            if lam != 0:
#                nc = np.size(RR,1)
#                print(m,n,(n+1)*m+n, np.size(RR,1))
                nc = (n+1)*(m+1)+n
                yy1 = np.concatenate((yy, np.zeros(nc)),0)
                RR1 = np.concatenate((RR, lam*linalg.norm(RR)*np.eye(nc)), 0)
                lsans = linalg.lstsq(RR1, yy1)
            else:
                lsans = linalg.lstsq(RR, yy)
            ht = lsans[0]
            ht = np.flip(ht,0)
            ff[fi,pidx,:] = ht[:(n+1)*m:(n+1)]
    #print(np.shape(RR))
    return ff

def kung_realization(mp, n, q=0):
    """
        Calulates the state-space realization `(a,b,c)` from Markov parameters
        using Kung's relization algorithm'

    Parameters
    ----------
    m : array of size(N,p,m) where N>=2n
        Array of Marov parameters m[i,j-1,k-1] hold row j
        and column k of the Markov parameter of sample i
    n : integer
        Model order state-space system.
    q : integer, optional
        Number of rows in Hankel matrix. The default is n+1.

    Returns
    -------
    tuple (a,b,c).

    """
    if q==0:
        q=n+1
    N = np.size(mp, 0)
    p = np.size(mp, 1)
    m = np.size(mp, 2)
    ncol = N-q+1
    if ncol<n:
        print("n, q and N are not compatible. N>=q+n-1 must be satisfied")
        return False
    H = np.empty((p*q, ncol*m), dtype=type(mp[0,0,0]))
    for j in range(ncol):
        for i in range(q):
            H[i*p:(i+1)*p, j*m:(j+1)*m] =mp[i+j, :, :]
    u, s, vh = linalg.svd(H, full_matrices=False, overwrite_a=True)
    c = u[:p, :n]
    lh = u[:p*(q-1), :n]
    rh = u[p:, :n]
    lsres = linalg.lstsq(lh, rh, overwrite_a=True, overwrite_b=True)
    a = lsres[0]
    b = np.dot(np.diag(s[:n]), vh[:n,:m])
    return a, b, c

def markov(sys, N):
    """
    Calculate markov parameters from sys = (a,b,c)

    Parameters
    ----------
    sys : tuple
        (a,b,c) state-space matrices.
    N : integer
        numer of Markov parameters to generate.

    Returns
    -------
    mp : array
        mp[i,:,;] is Markov parameter C(A^i)B.
    """
    (a, b, c) = sys[:3]
    n = np.size(a, 0)
    m = np.size(b, 1)
    p = np.size(c, 0)
    mp = np.empty((N, p, m), dtype=type(a[0,0]))
    aa = np.eye(n, dtype=type(a[0,0]))
    for i in range(N):
        mp[i, :, :] = np.dot(c, np.dot(aa, b))
        aa = np.dot(aa, a)
    return mp

def make_sys_real(sys):
    """
    Convert realization sys into a real-valued realization

    Parameters
    ----------
    sys : tuple
        sys = (a, b, c) or sys =  (a, b, c, d)

    Returns
    -------
    sysr : tuple
        sysr = (ar, br, cr, dr) the realization with real valued matrices

    """
    a, b, c = sys[:3]
    n = np.size(a,0)
    mp = markov(sys, 2*n )
    mpr = np.real(mp)
    a, b, c = kung_realization(mpr, n)
    if len(sys)==3:
        return (a, b, c)
    else:
        return (a, b, c, np.real(sys[3]))

def make_obs_real(a,c):
    """
    Convert (a,c) into real-valued matrices by approximating a real valued
    observability range space to the original one

    Parameters
    ----------
    a : matrix
    b : matrix

    Returns
    -------
    ar, br  : tuple
        the real valued matrices

    """
    n = np.size(a,0)
    p = np.size(c,0)
    obs = np.empty((p*(n+1), n), dtype='complex')
    obs[0:p, :] = c
    for i in range(n):
        obs[p*(i+1):p*(i+2), :] = np.dot(obs[p*i:p*(i+1), :], a)
    obsr = np.concatenate((np.real(obs), np.imag(obs)), 1)
    u, s, vh = linalg.svd(obsr, full_matrices=False, overwrite_a=True)
    c = u[:p, :n]
    lh = u[:p*n, :n]
    rh = u[p:, :n]
    lsres = linalg.lstsq(lh, rh, overwrite_a=True, overwrite_b=True)
    a = lsres[0]
    return a, c

def ls_estim_cd(ffdata, z, a, b, dtype='float', estimd=True):
    """
    Estimates the c and d matrices given a, b matrices and frequency function data 'ffdata'.

    Calulates the c and d matrices for a linear dynamic system given the a and b
    matrices and samples of frequency function data. It solves

    min_{c,d} sum_i || d + c*inv(z[i]*I-a)*b - ffdata[i,:,:] ||^2_F

    if estimd == False

    min_{c} sum_i || c*inv(z[i]*I-a)*b - ffdata[i,:,:] ||^2_F

    if dtype='float' a real valued solution is calulated. if dtype='complex'
    the solution is complex valued

    Parameters
    ----------
    ffdata : matrix_like
        frequency data packed in a matrix. fftata[i,:,:] is the frequency function
        matrix corresponding to sample i
    z : matrix_like
        vector with the frequecy points to evaluate, i.e. for DT z = exp(j*w) where w is frequencies in
        radians per sample
    a : matrix_like
        the a matrix
    b : matrix_like
        the b matrix
    dtype: optional
    estimd: optional
        data type of model either 'float' or 'complex'
        if set to False no d matrix is esimated and a zero d matrix is returned

    Returns
    -------

    c : matrix_like
        the LS-optimal c matrix
    d : matrix_like
        the LS-optimal d matrix

    """
    m = np.size(b, 1)
    nw = np.size(ffdata, 0)
    fk = ltifr_slow(a, b, z)
    fkstack = np.concatenate(fk, 1) #regressor for c matrix
    if estimd:
        dreg = np.tile(np.eye(m), (1, nw)) #regressor for d matrix
        R = np.concatenate((dreg, fkstack)) #complete regressor
    else:
        R = fkstack
    RH = np.concatenate(ffdata, 1) #right hand side in LS
    if dtype == 'float':
        RR = np.concatenate([np.real(R), np.imag(R)], 1)
        RRH = np.concatenate([np.real(RH), np.imag(RH)], 1)
        lsans = linalg.lstsq(RR.T, RRH.T, overwrite_a=True, overwrite_b=True)
    else:
        lsans = linalg.lstsq(R.T, RH.T, overwrite_a=True, overwrite_b=True)
    dc = lsans[0].T
    if estimd:
        return dc[:, m:], dc[:, :m]              #return c and d separatly
    else:
        return dc, np.zeros((np.size(dc, 0), m), dtype)

def ls_estim_bd(ffdata, z, a, c, dtype='float', estimd=True):
    """
    Estimates the b and d matrices given a, c matrices and frequency function data 'ffdata'.

    Calulates the b and d matrices for a linear dynamic system given the a and b
    matrices and samples of frequency function data. It solves

    min_{b,d} sum_i || d + c*inv(z[i]*I-a)*b - ffdata[i,:,:] ||^2_F

    if estimd == False

    min_{b} sum_i || c*inv(z[i]*I-a)*b - ffdata[i,:,:] ||^2_F

    if dtype='float' a real valued solution is calulated. if dtype='complex'
    the solution is complex valued

    Parameters
    ----------
    ffdata: matrix_like
        frequency data packed in a matrix. fftata[i,:,:] is the frequency function
        matrix corresponding to sample i
    z : matrix_like
        vector with the frequecy points to evaluate, i.e. for DT z = exp(j*w) where w is frequencies in
        radians per sample
    a: matrix_like
        the a matrix
    c: matrix_like
        the c matrix
    dtype: optional
        data type of model either 'float' or 'complex'
    estimd: optional
        if set to False no d matrix is esimated and a zero d matrix is returned

    Returns
    -------

    b : matrix_like
        the LS-optimal b matrix
    d : matrix_like
        the LS-optimal d matrix
    """

    fdt = transp_ffdata(ffdata)
    bt, dt = ls_estim_cd(fdt, z, a.T, c.T, dtype, estimd)
    return bt.T, dt.T

def transp_ffdata(ffdata):
    """Transposes ffdata (changes inputs and outputs) """
    m = np.size(ffdata, 2)
    p = np.size(ffdata, 1)
    nw = np.size(ffdata, 0)
    fdt = np.empty([nw, m, p], dtype='complex')
    for idx in range(nw):
        fdt[idx, :, :] = ffdata[idx, :, :].T
    return fdt

def fresp_slow(z, a, b, c, d):
    """Frequency response of ss-model (a,b,c,d)
    calulated as fresp[i,:,:] = d+c*inv(I*z[i]-a)*b (slow version)"""
    n = np.size(a, 0)
    m = np.size(b, 1)
    p = np.size(c, 0)
    nw = np.size(z, 0)
    frsp = np.empty([nw, p, m], dtype='complex')
    for widx in range(nw):
        aa = np.eye(n) * z[widx] - a
        frsp[widx, :, :] = np.linalg.multi_dot([c, np.linalg.solve(aa, b)]) + d
    return frsp

def fresp(z, a, b, c, d, noWarning=False):
    """Frequency response of ss-model (a,b,c,d)
    calulated as fresp[i,:,:] = d+c*inv(I*z[i]-a)*b (fast version)"""
    n = np.size(a, 0)
    m = np.size(b, 1)
    p = np.size(c, 0)
    nw = np.size(z, 0)
    lam, t = linalg.eig(a)
    if np.linalg.matrix_rank(t) < n:
        if  not noWarning:
            print("Matrix a defective. Eigenvectors do not form a basis. Using slow mode.")
        return fresp_slow(z, a, b, c, d)
    else:
        it = linalg.inv(t)
        bb = np.dot(it, b)
        cc = np.dot(c, t)
        frsp = np.empty([nw, p, m], dtype='complex')
        for widx in range(nw):
            da = np.ones(n) * z[widx] - lam
            frsp[widx, :, :] = np.linalg.multi_dot([cc, np.diag(1/da), bb]) + d
        return frsp

def ltifr_slow(a, b, z):
    """Calculates the (nz, n, m) size frequency kernel
      fkern[i,:,:] = inv(z[i]*eye(n)-a)*b (slow version)
    """
    n = np.size(a, 0)
    m = np.size(b, 1)
    nw = len(z)
    fkern = np.empty([nw, n, m], dtype='complex')
    for widx in range(nw):
        fkern[widx, :, :] = np.linalg.solve(np.eye(n) * z[widx] - a, b)
    return fkern

def ltifr(a, b, z, noWarning=False):
    """Calculates the (nz,n,m) size frequency kernel
        fkern[i,:,:] = inv(z[i]*eye(n)-a)*b
    """
    n = np.size(a, 0)
    m = np.size(b, 1)
    nw = len(z)
    lam, t = linalg.eig(a)
    if np.linalg.matrix_rank(t) < n:
        if not noWarning:
            print("Matrix a defective. Eigenvectors do not form a basis. Using slow mode.")
        return ltifr_slow(a, b, z)
    else:
        it = linalg.inv(t)
        bb = np.dot(it, b)
        fkern = np.empty([nw, n, m], dtype='complex')
        for widx in range(nw):
            da = np.ones(n) * z[widx] - lam
            fkern[widx, :, :] = np.linalg.multi_dot([t, np.diag(1/da), bb])
        return fkern

def ffsid_old(w, ffdata, n, q, dtype='float', estimd=True, CT=False, T=1):
    """Estimate a state-space model (a,b,c,d) from frequency function data

    Parameters
    ==========
    w : array_like
        vector of frequencies in rad/sample [-pi,pi] or [0, 2pi]
        if CT = True unit it radians/s (-inf, +inf)
    ffdata : array_like
        matrix of frequency function data. ffdata[i,:,:] is the frequency response matrix
        at frequency w[i]
    n : integer
        the model order of the ss-model (a,b,c,d), i.e. a is a size (n x n) matrix
    q : integer
        the numer of block rows used in the intermediate matrix. Must satisfy q>n
    dtype : data type, optional
        if dtype = 'float' a real valued solution (a,b,c,d) is returned.
        if dtype = 'complex' a complex valued solution (a,b,c,d) is returned.
    estimd: optional
        if set to False no d matrix is esimated and a zero d matrix is returned
    CT : if set to true a continuous time (CT) model is esimated
         if set to false a discrete time (DT) model is esimated (default)
    T :  a frequency scaling factor for the bilinear transformation used when CT=True.
        Default is 1. If CT=False parameter T is disregarded

    Returns
    =======
    a  : matrix_like
        the estimated a matrix
    b  : matrix_like
        the estimated b matrix
    c  : matrix_like
        the estimated c matrix
    d  : matrix_like
        the estimated d matrix (or zero matrix if estimd=False)
    s  : matrix_like
        a vector of the singular values


    """
    if CT:
        ad, bd, cd, dd, s  = ffsid(cf2df(w,T), ffdata, n, q, dtype, estimd=True, CT=False, T=T)
        a, be, ce, de = bilinear_d2c((ad, bd, cd, dd), T)
        if not estimd:
            be, de = ls_estim_bd(ffdata, 1j*w, a, ce, dtype, estimd)
            ce, de = ls_estim_cd(ffdata, 1j*w, a, be, dtype, estimd)
            be, de = ls_estim_bd(ffdata, 1j*w, a, ce, dtype, estimd)
            ce, de = ls_estim_cd(ffdata, 1j*w, a, be, dtype, estimd)
        return a, be, ce, de, s

    nwfd, p, m = np.shape(ffdata)
    nw = np.size(w, 0)
    z = np.exp(1j*w)
    if q < n+1:
        print("Error: q must satidfy q>n.")
        return False
    if nw != nwfd:
        print("Error: w and ffdata sizes does not match!")
        return False
    y = np.empty([p*q, nw*m], dtype='complex')
    u = np.empty([m*q, nw*m], dtype='complex')
    for widx in range(nw):
        y[:p, widx*m:(widx+1)*m] = ffdata[widx, :, :]
        u[:m, widx*m:(widx+1)*m] = np.eye(m)
        zx = z[widx]
        for qidx in range(q)[1:]:
            y[qidx*p:(qidx+1)*p, widx*m:(widx+1)*m] = zx*ffdata[widx, :, :]
            u[qidx*m:(qidx+1)*m, widx*m:(widx+1)*m] = zx*np.eye(m)
            zx *= z[widx]
    if dtype == 'float':
        hU = np.concatenate((np.real(u), np.imag(u)), 1)
        hY = np.concatenate((np.real(y), np.imag(y)), 1)
    else:
        hU = u
        hY = y
    h = np.concatenate((hU, hY))
    r = linalg.qr(h.T, mode='r', overwrite_a=True)  #Calulates the projection
    r = r[0].T
    r22 = r[m*q:, m*q:]
    u, s, vh = linalg.svd(r22, full_matrices=False, overwrite_a=True)
    c = u[:p, :n]
    lh = u[:p*(q-1), :n]
    rh = u[p:, :n]
    lsres = linalg.lstsq(lh, rh, overwrite_a=True, overwrite_b=True)
    a = lsres[0]
    be, de = ls_estim_bd(ffdata, z, a, c, dtype, estimd)
    ce, de = ls_estim_cd(ffdata, z, a, be, dtype, estimd)
    be, de = ls_estim_bd(ffdata, z, a, ce, dtype, estimd)
    ce, de = ls_estim_cd(ffdata, z, a, be, dtype, estimd)
    return a, be, ce, de, s


def ffsid(w, ffdata, n, q, dtype='float', estimd=True, CT=False, T=1):
    """Estimate a state-space model (a,b,c,d) from frequency function data

    Parameters
    ==========
    w : array_like
        vector of frequencies in rad/sample [-pi,pi] or [0, 2pi]
        if CT = True unit in radians/s (-inf, +inf)
    ffdata : array_like
        matrix of frequency function data. ffdata[i,:,:] is the frequency response matrix
        at frequency w[i]
    n : integer
        the model order of the ss-model (a,b,c,d), i.e. a is a size (n x n) matrix
    q : integer
        the numer of block rows used in the intermediate matrix. Must satisfy q>n
    dtype : data type, optional
        if dtype = 'float' a real valued solution (a,b,c,d) is returned.
        if dtype = 'complex' a complex valued solution (a,b,c,d) is returned.
    estimd: optional
        if set to False no d matrix is esimated and a zero d matrix is returned
    CT: if set to true a continuous time (CT) model is esimated
         if set to false a discrete time (DT) model is esimated (default)
    T :  a frequency scaling factor for the bilinear transformation used when CT=True.
        Default is 1. If CT=False parameter T is disregarded

    Returns
    =======
    a  : matrix_like
        the estimated a matrix
    b  : matrix_like
        the estimated b matrix
    c  : matrix_like
        the estimated c matrix
    d  : matrix_like
        the estimated d matrix (or zero matrix if estimd=False)
    s  : matrix_like
        a vector of the singular values


    """
    if CT:
        ad, bd, cd, dd, s  = ffsid(cf2df(w,T), ffdata, n, q, dtype, estimd=True, CT=False, T=T)
        a, be, ce, de = bilinear_d2c((ad, bd, cd, dd), T)
        if not estimd:
            be, de = ls_estim_bd(ffdata, 1j*w, a, ce, dtype, estimd)
            ce, de = ls_estim_cd(ffdata, 1j*w, a, be, dtype, estimd)
            be, de = ls_estim_bd(ffdata, 1j*w, a, ce, dtype, estimd)
            ce, de = ls_estim_cd(ffdata, 1j*w, a, be, dtype, estimd)
        return a, be, ce, de, s

    z = np.exp(1j*w)
    return gffsid(z, ffdata, n, q, dtype=dtype, estimd=estimd)

def gffsid(z, ffdata, n, q, dtype='float', estimd=True):
    """Estimate a state-space model (a,b,c,d) from general frequency function data
    by mininizing the Frobenius norm
        sum_i   || ffdata[i,:,:] - d - c*inv(z[i]*eye(n)-A)*b ||^2_F

    Parameters
    ==========
    z : array_like
        vector of complex numbers
    ffdata : array_like
        matrix of frequency function data. ffdata[i,:,:] is the frequency response matrix
        at complex point z[i]
    n : integer
        the model order of the ss-model (a,b,c,d), i.e. a is a size (n x n) matrix
    q : integer
        the numer of block rows used in the intermediate matrix. Must satisfy q>n
    dtype : data type, optional
        if dtype = 'float' a real valued solution (a,b,c,d) is returned.
        if dtype = 'complex' a complex valued solution (a,b,c,d) is returned.
    estimd: optional
        if set to False no d matrix is esimated and a zero d matrix is returned

    Returns
    =======
    a  : matrix_like
        the estimated a matrix
    b  : matrix_like
        the estimated b matrix
    c  : matrix_like
        the estimated c matrix
    d  : matrix_like
        the estimated d matrix (or zero matrix if estimd=False)
    s  : matrix_like
        a vector of the singular values

    """

    nwfd, p, m = np.shape(ffdata)
    nw = np.size(z, 0)
    if q < n+1:
        print("Error: q must satisfy q>n.")
        return False
    if nw != nwfd:
        print("Error: w and ffdata sizes does not match!")
        return False
    y = np.empty([p*q, nw*m], dtype='complex')
    if estimd:
        u = np.empty([m*q, nw*m], dtype='complex')
    else:
        u = np.empty([m*(q-1), nw*m], dtype='complex')


    for widx in range(nw):
        y[:p, widx*m:(widx+1)*m] = ffdata[widx, :, :]
        u[:m, widx*m:(widx+1)*m] = np.eye(m)
        zx = z[widx]
        for qidx in range(q)[1:]:
            y[qidx*p:(qidx+1)*p, widx*m:(widx+1)*m] = zx*ffdata[widx, :, :]
            if estimd or qidx<q-1:
                u[qidx*m:(qidx+1)*m, widx*m:(widx+1)*m] = zx*np.eye(m)
            zx *= z[widx]
    if dtype == 'float':
        hU = np.concatenate((np.real(u), np.imag(u)), 1)
        hY = np.concatenate((np.real(y), np.imag(y)), 1)
    else:
        hU = u
        hY = y
    h = np.concatenate((hU, hY))
    r = linalg.qr(h.T, mode='r', overwrite_a=True)  #Calulates the projection
    r = r[0].T
    if estimd:
        r22 = r[m*q:, m*q:]
    else:
        r22 = r[m*(q-1):, m*(q-1):]
    u, s, vh = linalg.svd(r22, full_matrices=False, overwrite_a=True)
    c = u[:p, :n]
    lh = u[:p*(q-1), :n]
    rh = u[p:, :n]
    lsres = linalg.lstsq(lh, rh, overwrite_a=True, overwrite_b=True)
    a = lsres[0]
    be, de = ls_estim_bd(ffdata, z, a, c, dtype, estimd)
    ce, de = ls_estim_cd(ffdata, z, a, be, dtype, estimd)
    be, de = ls_estim_bd(ffdata, z, a, ce, dtype, estimd)
    ce, de = ls_estim_cd(ffdata, z, a, be, dtype, estimd)
#    pip = np.eye(m*nw) - np.linalg.multi_dot( [u.T.conj(), np.linalg.inv( np.dot(u , u.T.conj())),  u])
#    x = np.dot(y, pip)
#    pip, u, R, Gproj = project_fdshape(x, z, p, m, q)
#    print(np.linalg.norm(Gproj-ffdata))
    return a, be, ce, de, s


def fdsid(fddata, n, q, estTrans=True, dtype='float', estimd=True, CT=False, T=1, W=np.empty(0)):
    """
    Estimate a DT or CT state-space model from I/O frequency data

    Determines the (a,b,c,d,xt) parametrers such that (DT case)
     sum_i   || y[i,:] - d*u[i, :] + c*inv(z[i]*eye(n)-A)*[b,  xt]* [u[i, :]; z[i]]||^2_F
    is small where z = np.exp(1j*w)
    and CT Case
     sum_i   ||  y[i,:] - d*u[i, :] + c*inv(1j*w[i]*eye(n)-A)*b* u[i, :] ||^2_F


    Parameters
    ==========
    fddata : tuple
        fddata[0] = w
            a vector of frequencies
            in radians/sample
        fddata[1] = y
            a matrix of the output transform data where y[i,:] is the DFT of the
            time domain output signal corresponding to frequency w[i]
        fddata[2] = u
            a matrix of the input transform data where u[i,:] is the DFT of the
            time domain input signal corresponding to frequency w[i]
    n : integer
        the model order of the ss-model
    q : integer
        the numer of block rows used in the intermediate matrix. Must satisfy q>n
    estTrans : boolean, optional
        if True, a compensation for the transient term will be estimated (default)
    dtype : data type, optional
        if dtype = 'float' a real valued solution is returned.
        if dtype = 'complex' a complex valued solution is returned.
    estimd : boolean, optional
        if set to False no d matrix is esimated and a zero d matrix is returned
    CT : bool
        if True a CT model is estimated and estTrans is forced False.
    T :  a frequency scaling factor for the bilinear transformation used when CT=True.
        Default is 1. If CT=False parameter T is disregarded


    Returns
    =======
    a  : matrix_like
        the estimated a matrix
    b  : matrix_like
        the estimated b matrix
    c  : matrix_like
        the estimated c matrix
    d  : matrix_like
        the estimated d matrix
    xt : matrix_like
        vector of the transient compensation
    s  : matrix_like
        a vector of the singular values
    """
    w = fddata[0]
    yd = fddata[1]
    ud = fddata[2]
    if CT:
        estTrans=False
        ad, bd, cd, dd, xt, s  = fdsid((cf2df(w,T), yd, ud), n, q, estTrans=False, dtype=dtype, estimd=True, CT=False, T=T, W=W)
        a, b, c, d = bilinear_d2c((ad, bd, cd, dd), T)
        if not estimd:
            b, d, resid = fdestim_bd(1j*w, yd, ud, a, c, estTrans, dtype, estimd, w=W)
            c, d, resid = fdestim_cd(1j*w, yd, ud, a, b, 0, estTrans, dtype, estimd, w=W)
            b, d, resid = fdestim_bd(1j*w, yd, ud, a, c, estTrans, dtype, estimd, w=W)
            c, d, resid = fdestim_cd(1j*w, yd, ud, a, b, 0, estTrans, dtype, estimd, w=W )
            xt = np.zeros((n, 1), dtype)
        return a, b, c, d, xt, s

    return gfdsid((np.exp(1j*w), yd, ud), n, q, estTrans=estTrans, dtype=dtype, estimd=estimd, w=W)


def gfdsid(fddata, n, q, estTrans=True, dtype='float', estimd=True, w=np.empty(0) ):
    """
    Estimate a state-space model from I/O frequency data

    Determines the (a,b,c,d,xt) parametrers such that
     sum_i || y[i,:] - d*u[i, :] + c*inv(z[i]*eye(n)-A)*[b,  xt]* [u[i, :]; z[i]] ||^2_w[i,:,:]
     is minimized
    If estrTrans=False the following problem is solved
     sum_i ||y[i,:] - d*u[i, :] + c*inv(z[i]*eye(n)-A)*b * u[i, :] ||^2_w[i,:,:]
     is minimized

    The weighted norm  || x ||^2_w is defined as  || w * x ||^2 where w is a square matrix such that w*w^H
    is a positive definite matrix.
    If the noise on y[i,:].T is a zero mean rv with covariance r[i,:,:] a BLUE estimator will be obtained if
    w[i,:,:] is selected as linalg.cholesky(linalg.inv(r[i,:,:]])).T.conj()


    Parameters
    ==========
    fddata : tuple
        fddata[0] = z
            a vector of complex scalars
        fddata[1] = y
            a matrix of the output frequency data where y[i,:] corresponding to z[i]
        fddata[2] = u
            a matrix of the input frrequency data where u[i,:] corresponding to z[i]
    n : integer
        the model order of the ss-model
    q : integer
        the numer of block rows used in the intermediate matrix. Must satisfy q>n
    estTrans : boolean, optional
        if True, a compensation for the transient term will be estimated (default)
    dtype : data type, optional
        if dtype = 'float' a real valued solution is returned.
        if dtype = 'complex' a complex valued solution is returned.
    estimd : boolean, optional
        if set to False no d matrix is esimated and a zero d matrix is returned
    w: matrix like, optional
        w[i,:,:] is the weighting matrix for data sample i, See above.

    Returns
    =======
    a  : matrix_like
        the estimated a matrix
    b  : matrix_like
        the estimated b matrix
    c  : matrix_like
        the estimated c matrix
    d  : matrix_like
        the estimated d matrix
    xt : matrix_like
        vector of the transient compensation
    s  : matrix_like
        a vector of the singular values
    """
    z = fddata[0]
    yd = fddata[1]
    ud = fddata[2]

    nwy, p = np.shape(yd)
    nwu, m = np.shape(ud)
    if estTrans:
        ude = np.concatenate((ud, np.asmatrix(z).T), 1)
        me = m + 1
    else:
        ude = ud
        me = m
    nw = np.size(z, 0)
    if nw != nwy:
        print("Error: z and Y sizes does not match!")
        return False
    if nw != nwu:
        print("Error: z and U sizes does not match!")
        return False
    y = np.empty([p*q, nw], dtype='complex')
    if estimd:
        u = np.empty([me*q, nw], dtype='complex')
    else:
        u = np.empty([me*(q-1), nw], dtype='complex')
    wf = np.ones((nw,1))
    if w.size>0:
        for widx in range(nw):
            wf[widx] = np.trace(w[widx,:,:])
    for widx in range(nw):
        y[:p, widx] = wf[widx]*yd[widx, :]
        u[:me, widx] = wf[widx]*ude[widx, :]
        zx = z[widx]
        for qidx in range(q)[1:]:
            y[qidx*p:(qidx+1)*p, widx] = wf[widx]*zx*yd[widx, :]
            if estimd or qidx<q-1:
                u[qidx*me:(qidx+1)*me, widx] = wf[widx]*zx*ude[widx, :]
            zx *= z[widx]

    if dtype == 'float':
        hu = np.concatenate((np.real(u), np.imag(u)), 1)
        hy = np.concatenate((np.real(y), np.imag(y)), 1)
    else:
        hu = u
        hy = y
    h = np.concatenate((hu, hy))
    r = linalg.qr(h.T, mode='r', overwrite_a=True)  #Calulates the projection
    r = r[0].T
    if estimd:
        r22 = r[me*q:, me*q:]
    else:
        r22 = r[me*(q-1):, me*(q-1):]

    u, s, vh = linalg.svd(r22, full_matrices=False, overwrite_a=True)
    c = u[:p, :n]
    lh = u[:p*(q-1), :n]
    rh = u[p:, :n]
    lsres = linalg.lstsq(lh, rh, overwrite_a=True, overwrite_b=True)
    a = lsres[0]
    if estTrans:
        b, d, xt, resid = fdestim_bd(z, yd, ud, a, c, estTrans, dtype, estimd, w)
        c, d, resid = fdestim_cd(z, yd, ud, a, b, xt, estTrans, dtype, estimd, w)
        b, d, xt, resid = fdestim_bd(z, yd, ud, a, c, estTrans, dtype, estimd, w)
        c, d, resid = fdestim_cd(z, yd, ud, a, b, xt, estTrans, dtype, estimd, w)
    else:
        b, d, resid = fdestim_bd(z, yd, ud, a, c, estTrans, dtype, estimd, w)
        c, d, resid = fdestim_cd(z, yd, ud, a, b, 0, estTrans, dtype, estimd, w)
        b, d, resid = fdestim_bd(z, yd, ud, a, c, estTrans, dtype, estimd, w)
        c, d, resid = fdestim_cd(z, yd, ud, a, b, 0, estTrans, dtype, estimd, w)
        xt = np.zeros((n, 1), dtype)
    return a, b, c, d, xt, s


def ltifd_slow(a, b, u, z):
    """Calculates the (nz, n) size frequency kernel with input u
        fkern[:, i] = inv(eye(n)*z[i] - a)*b*u[i, :]
        (slow version)
        """
    n = np.size(a, 0)
    nw = len(z)
    fkern = np.empty([n, nw], dtype='complex')
    for widx in range(nw):
        fkern[:, widx] = linalg.solve(np.eye(n) * z[widx] - a, np.dot(b, u[widx, :]))
    return fkern

def ltifd(a, b, u, z, noWarning=False):
    """Calculates the nz x n frequency kernel with input u
        fkern[:, i] = inv(eye(n)*z[i] - a)*b*u[i, :]
        (fast version)"""
    n = np.size(a, 0)
    nw = len(z)
    lam, t = linalg.eig(a)
    if np.linalg.matrix_rank(t) < n:
        if not noWarning:
            print("Matrix a defective. Eigenvectors do not form a basis. Using slow mode.")
        return ltifd_slow(a, b, u, z)
    else:
        it = linalg.inv(t)
        bb = np.dot(it, b)
        fkern = np.empty([n, nw], dtype='complex')
        for widx in range(nw):
            da = np.ones(n) * z[widx] - lam
            fkern[:, widx] = np.linalg.multi_dot([t, np.diag(1/da), np.dot(bb, u[widx, :])])
        return fkern


def fdestim_cd(z, yd, ud, a, b, xt=0, estTrans=False, dtype='float', estimd=True, w = np.empty(0)):
    """Estimate c and d matrices given z, yd, ud and a, b, and (optionally xt) matrices

    Calulates the c and d matrices for a state-space representation of a rational function
    given the a and b matrices and samples of the rational matrix function data in the form of
    output vectors yd[i,:] and input vectors ud[i,:] corresponding to the rational function
    evaluated at complex value z[i]. It solves

    if estimd=True and estTrans=True
    min_{c,d} sum_i || ([d 0] + c*inv(z[i]*I-a)*[b xt])*[ud[i,:]; z[i]]  - yd[i,:] |^2_w[i,:,:]

    if estimd=False and estTrans=True
    min_{c} sum_i|| (c*inv(z[i]*I-a)*[b xt])*[ud[i,:]; z[i]]  - yd[i,:] |^2_w[i,:,:]

    if estimd=True and estTrans=False
    min_{c,d} sum_i || (d+ c*inv(z[i]*I-a)*b)*ud[i,:]  - yd[i,:] |^2_w[i,:,:]

    if estimd=False and estTrans=False
    min_{c} sum_i|| (c*inv(z[i]*I-a)*b)*ud[i,:]  - yd[i,:] |^2_w[i,:,:]

    The weighted norm  || x ||^2_w is defined as  || w * x ||^2 where w is a square matrix such that w*w^H
    is a positive definite matrix.
    If the noise on y[i,:].T is a zero mean rv with covariance r[i,:,:] a BLUE estimator will be obtained if
    w[i,:,:] is selected as linalg.cholesky(linalg.inv(r[i,:,:]])).T.conj()


    if dtype='float' a real valued solution is calulated. if dtype='complex'
    the solution is complex valued

    Parameters
    ----------
    yd : matrix_like
        frequency data packed in a matrix. yd[i,:] is the output vector
        corresponding to sample i
    ud : matrix_like
        frequency data packed in a matrix. ud[i,:] is the output vector
        corresponding to sample i
    z : matrix_like
        vector with the complex data z[i].
    a : matrix_like
        the a matrix
    b : matrix_like
        the b matrix
    dtype : optional
        data type of model either 'float' or 'complex'
    estimd : boolean, optional
        if set to False no d matrix is esimated and a zero d matrix is returned
      w: matrix like, optional
        w[i,:,:] is the weighting matrix for data sample i, See above.

    Returns
    -------

    c : matrix_like
        the LS-optimal c matrix
    d : matrix_like
        the LS-optimal d matrix

    """
    p = np.shape(yd)[1]
    m = np.shape(ud)[1]
    n = np.shape(a)[0]
    nz = np.size(z, 0)
    if estTrans == True:
        ude = np.concatenate((ud, np.array(z).reshape(nz, 1)), 1)
        be = np.concatenate((b, xt.reshape(n, 1)), 1)
    else:
        ude = ud
        be = b
    fkern = ltifd(a, be, ude, z)
    if estimd:
        r = np.empty((n+m, nz), dtype='complex')
        r[:n, :] = fkern
        r[n:, :] = np.transpose(ud)
    else:
        r = fkern
    if w.size>0:
        nr = np.size(r,0)
        rw = np.empty((p*nz, nr*p), dtype='complex')
        ydw = np.empty((p*nz), dtype='complex')
        for zidx in range(nz): #Vectorize data and apply pre-whitening filter
            ydw[p*zidx:p*(zidx+1)] = np.matmul(w[zidx,:,:], yd[zidx,:])
            rw[p*zidx:p*(zidx+1),:] = np.kron(r[:,zidx].T, w[zidx,:,:])

        if dtype == 'float':
            lh = np.concatenate([np.real(ydw), np.imag(ydw)], 0)
            rh = np.concatenate([np.real(rw), np.imag(rw)], 0)
        else:
            lh = ydw
            rh = rw
        lsres = linalg.lstsq(rh, lh)
        vecCD = lsres[0]
        c = np.reshape(vecCD[:n*p], (p, n), order='F')
        if estimd:
            return c, np.reshape(vecCD[n*p:], (p, m), order='F'), lsres[1]
        else:
            return c , np.zeros((p, m), dtype), lsres[1] # Return c and d and residuals
    else:
        if dtype == 'float':
            rh = np.concatenate([np.real(r), np.imag(r)], 1)
            lh = np.concatenate([np.real(yd), np.imag(yd)], 0)
        else:
            lh = yd
            rh = r
        lsres = linalg.lstsq(rh.T, lh)
        cd = lsres[0].T
        if estimd:
            return cd[:, :n], cd[:, n:], lsres[1] # Return c and d and residuals
        else:
            return cd, np.zeros((p, m), dtype), lsres[1] # Return c and d and residuals

        if estimd:
            return cd[:, :n], cd[:, n:], lsres[1] # Return c and d and residuals
        else:
            return cd, np.zeros((p, m), dtype), lsres[1] # Return c and d and residuals

def fdestim_bd(z, yd, ud, a, c, estTrans=False, dtype='float', estimd=True, w = np.empty(0)):
    """Estimate b and d (and optionally xt) matrices given z, yd, ud and a, c matrices

    Calulates the b and d (and optionally xt) matrices for a state-space representation of
    a rational function
    given the a and b matrices and samples of the rational matrix function data in the form of
    output vectors yd[i,:] and input vectors ud[i,:] corresponding to the rational function
    evaluated at complex value z[i]. It solves

    if estimd=True and estTrans=True
    min_{b,d,xt} sum_i || ([d 0] + c*inv(z[i]*I-a)*[b xt])[ud[i,:]; z[i]]  - yd[i,:] ||^2_w[i,:,:]

    if estimd=False and estTrans=True
    min_{b,xt} sum_i|| (c*inv(z[i]*I-a)*[b xt])[ud[i,:]; z[i]]  - yd[i,:] ||^2_w[i,:,:]

    if estimd=True and estTrans=False
    min_{b,d} sum_i || (d+ c*inv(z[i]*I-a)*b)ud[i,:]  - yd[i,:] ||^2_w[i,:,:]

    if estimd=False and estTrans=False
    min_{b} sum_i|| (c*inv(z[i]*I-a)*b)ud[i,:]  - yd[i,:] ||^2_w[i,:,:]

    The weighted norm  || x ||^2_w is defined as  || w * x ||^2 where w is a square matrix such that w*w^H
    is a positive definite matrix.
    If the noise on y[i,:].T is zero mean rv with covariance r[i,:,:] a BLUE estimator will be obtained if
    w[i,:,:] is selected as linalg.cholesky(linalg.inv(r[i,:,:]])).T.conj()

    if dtype='float' a real valued solution is calulated (default). if dtype='complex'
    the solution is complex valued

    Parameters
    ----------
    z : matrix_like
        vector with the rational function evaluation points z[i].
    yd : matrix like
        output frequency data yd[i,:]
    ud : matrix like
        input frequency data ud[i,:]
    a : matrix_like
        the a matrix
    c : matrix_like
        the c matrix
    estTrans : boolean, optional
        if set to True also an xt vector will be estimated capturing the transient effect
    dtype : optional
        data type of model either 'float' or 'complex'
    estimd : boolean, optional
        if set to False no d matrix is esimated and a zero d matrix is returned
    w: matrix like, optional
        w[i,:,:] is the weighting matrix for data sample i, See above.
    Returns
    -------

    c : matrix_like
        the LS-optimal c matrix
    d : matrix_like
        the LS-optimal d matrix


    """
    p = np.shape(yd)[1]
    m = np.shape(ud)[1]
    n = np.shape(a)[0]
    nz = np.size(z, 0)
    if w.size>0:
        ydw = np.empty((nz,p),dtype='complex')
        for zidx in range(nz):
            ydw[zidx,:] = np.matmul(w[zidx,:,:], yd[zidx,:])
    else:
        ydw = yd
    if estTrans == True:
        ude = np.concatenate((ud, np.array(z).reshape(nz, 1)), 1)
 #       ude = np.concatenate((ud, np.ones((NwU, 1))), 1)
        me = m + 1
    else:
        ude = ud
        me = m

    fkern = ltifr(a.T, c.T, z).T
    if estimd:
        r = np.empty([nz*p, me*n+m*p], dtype='complex') #room for B xt and D
        if w.size>0:
            for zidx in range(nz):
                for midx in range(me):
                    r[p*zidx:p*(zidx+1), n*midx:n*(midx+1)] = np.matmul(
                            w[zidx,:,:], ude[zidx, midx] * fkern[:, :, zidx])
                for midx in range(m):
                    r[p*zidx:p*(zidx+1), me*n+p*midx:me*n+p*(midx+1)] = np.matmul(
                            w[zidx,:,:], ud[zidx, midx] * np.eye(p))
        else:
            for zidx in range(nz):
                for midx in range(me):
                    r[p*zidx:p*(zidx+1), n*midx:n*(midx+1)] = ude[zidx, midx] * fkern[:, :, zidx]
                for midx in range(m):
                    r[p*zidx:p*(zidx+1), me*n+p*midx:me*n+p*(midx+1)] = ud[zidx, midx] * np.eye(p)
    else:
        r = np.empty([nz*p, me*n], dtype='complex') #room for B xt
        if w.size>0:
            for zidx in range(nz):
                for midx in range(me):
                    r[p*zidx:p*(zidx+1), n*midx:n*(midx+1)] = np.matmul(
                            w[zidx,:,:], ude[zidx, midx] * fkern[:, :, zidx])
        else:
            for zidx in range(nz):
                for midx in range(me):
                    r[p*zidx:p*(zidx+1), n*midx:n*(midx+1)] = ude[zidx, midx] * fkern[:, :, zidx]
    if dtype == 'float':
        rh = np.concatenate([np.real(r), np.imag(r)], 0)
        lh = np.concatenate([np.real(ydw.flatten()), np.imag(ydw.flatten())], 0)
    else:
        rh = r
        lh = ydw.flatten()
    lsres = linalg.lstsq(rh, lh)
    vecBD = lsres[0]
    vecB = vecBD[:n*me]
    if estimd:
        vecD = vecBD[n*me:]
    else:
        vecD = np.zeros(m*p)
    b = np.empty([n, m], dtype=dtype)
    d = np.empty([p, m], dtype=dtype)
    for midx in range(m):
        b[:, midx] = vecB[n*midx:n*(midx+1)]
        d[:, midx] = vecD[p*midx:p*(midx+1)]
    if estTrans:
        xt = vecB[n*m:n*(m+1)]
        return b, d, xt, lsres[1]
    else:
        return b, d, lsres[1]


def ffdata2fddata(ffdata, w):
    """Converts ffdata to fddata

    Converts frequency function data (ffdata) to input/output data format (fddata).

    Parameters
    ==========
    ffdata : array_like
        frequency function data in the format such that `ffdata[widx,:,:]`
        corresponds to the frequency function matrix of size `(p,m)` at frequency index `widx` at a total number of frequencies `nw`
    w : array_like
        array with the corresponding frequencies in radians per sample.

    Returns
    =======
    u : array_like
        Fourier transform of input of size `(m*nw,m)`
    y : array_like
        Fourier transform of output of size `(m*nw,p)`
    wn : array_like
        frequency vector of length `m*nw`
    """
    nw, p, m = np.shape(ffdata)
    u = np.empty([nw*m, m], dtype='complex')
    y = np.empty([nw*m, p], dtype='complex')
    wn = np.empty((nw*m))
    b0 = np.eye(m)
    for midx in range(m):
        y[midx*nw:(midx+1)*nw, :] = ffdata[:, :, midx]
        u[midx*nw:(midx+1)*nw, :] = np.tile(b0[midx, :], (nw, 1))
        wn[midx*nw:(midx+1)*nw] = w[:]
    return u, y, wn

def ltitr(a, b, u, x0=0, dtype='float'):
    """Calculates the time domain input to state respone

    Calculates the time domain state response
    x[i+1,:] = np.dot(a,x[i,:]) + np.dot(b,u[i,:])

    Parmeters
    =========
    a : array_like
        a square matrix of size `(n,n)`
    b : array_like
        a matrix of size `(n,m)`
    u : array_like
        an array of input vectors such that `u[i,:]` is
        the input vector of size `m` at time index `i`.
        The array has size `(N,m)`.
    x0 : array_like
        intial vector of size `n`, i.e. `x[0,:]=x0`. Default value is the zero vector.

    Returns
    =======
    x : array_like
        the resulting state-sequence of size `(N,n)`
        x[k,:] is the state at sample k
    """
    n = np.shape(a)[0]
    N, m = np.shape(u)
    x = np.empty((N, n), dtype)
    if x0 == 0:
        x0 = np.zeros((n), dtype)
    x[0, :] = x0
    for nidx in range(N-1):
        x[nidx+1, :] = np.dot(a, x[nidx, :]) + np.dot(b, u[nidx, :])
    return x

def lsim(sys, u, x0=0, dtype='float'):
    """Calculates the time-domain output given input sequence and state-space model

    Calculates the time domain state response

    x[i+1,:] = np.dot(a,x[i,:]) + np.dot(b,u[i,:])

    y[i,:] = np.dot(c,x[i,:]) + np.dot(d,u[i,:])

    Parmeters
    =========
    sys : tuple
        sys = (a, b, c, d) or
        sys = (a, b, c)
        where
        a square matrix of size (n,n), b is
        a matrix of size (n,m), c is a matrix of size (p,n)
        and (optionally) d is a matrix of size (p,m).

    u : array_like
        an array of input vectors such that u[i,:] is
        the input vector of size m at time index i.
        The array has size (N,m).

    x0 : array_like (optional)
        intial vector of size n, i.e. x[0,:]=x0. Default value is the zero vector.

    Returns
    =======
    y : array_like
        the resulting output sequence of size (N,p)   """

    nn = len(sys)
    if nn == 3:
        a, b, c = sys
        p, nc = np.shape(c)
        nr, m = np.shape(b)
        d = np.zeros(p, m)
    elif nn == 4:
        a, b, c, d = sys
        p, nc = np.shape(c)
        nr, m = np.shape(b)
    else:
        print("lsim: Incorrect number of matrices in sys.")
        return False
    x = ltitr(a, b, u, x0, dtype)
    nu = np.size(u, 0)
    y = np.empty((nu, p), dtype)
    for idx in range(nu):
        y[idx, :] = np.dot(c, x[idx, :]) + np.dot(d, u[idx, :])
    return y

def fdsim(sys, u, z, xt=np.empty(0)):
    """Calculates the output given input and state-space model in Fourier domain"""
    nwu, m = np.shape(u)
    nn = len(sys)
    if nn == 3:
        a, b, c = sys
        p, nc = np.shape(c)
        nr, m = np.shape(b)
        d = np.zeros(p, m)
    elif nn == 4:
        a, b, c, d = sys
        p, nc = np.shape(c)
        nr, m = np.shape(b)
    else:
        print("fdsim: Incorrect number of matrices in sys.")
        return False
    y = np.empty((nwu, p), dtype='complex')
    if np.size(xt) > 0:
        ue = np.concatenate((u, z.reshape(nwu, 1)), 1)
        be = np.concatenate((b, xt.reshape(nr, 1)), 1)
        x = ltifd(a, be, ue, z)
    else:
        x = ltifd(a, b, u, z)
    for widx in range(nwu):
        y[widx, :] = np.dot(c, x[:, widx]) + np.dot(d, u[widx, :])
    return y

def bilinear_d2c(sys, T=1):
    """ Calculates the bilinear transformation D->C for ss-system sys """
    a, b, c, d = sys
    n =  np.shape(a)[0]
    ainv = np.linalg.inv(np.eye(n)+a)
    ac = np.dot(ainv, a-np.eye(n))*2/T
    bc = np.dot(ainv, b)*2/np.sqrt(T)
    cc = np.dot(c, ainv)*2/np.sqrt(T)
    dc = d - np.linalg.multi_dot([c, ainv, b])
    return ac, bc, cc, dc

def bilinear_c2d(sys, T=1):
    """ Calculates the bilinear transformation C->D for ss-system sys """
    a, b, c, d = sys
    n =  np.shape(a)[0]
    ainv = np.linalg.inv(np.eye(n)*2/T-a)
    ad = np.dot(np.eye(n)*2/T+a, ainv )
    bd = np.dot(ainv, b)*2/np.sqrt(T)
    cd = np.dot(c, ainv)*2/np.sqrt(T)
    dd = d+np.linalg.multi_dot([c, ainv, b])
    return ad, bd, cd, dd

def cf2df(wc,T):
    """ Calculates the bilinear transformation frequency mapping C->D for frequency vector wc """
    return 2*np.arctan(wc*T/2)

def df2cf(wd,T):
    """ Calculates the bilinear transformation frequency mapping D->C for frequency vector wd """
    return 2*np.tan(wd/2)/T

def vec(a):
    """ Vectorize a matrix

    Input
    =====
    a : matrix (ndarray)

    Returns
    =======
        A vector with all columns in a concatenated to one long vector
    """
    return np.reshape(a,np.size(a), order='F')

def moebius(sys, par = (1, 0, 0, 1)):
    """ Calculates the bilinear transformation D->C for ss-system sys """
    a, b, c, d = sys
    alpha, beta, gamma, delta = par
    n =  np.shape(a)[0]
    ainv = np.linalg.inv(alpha*np.eye(n)-gamma*a)
    ac = np.dot(delta*a-beta*np.eye(n), ainv)
    bc = (alpha*delta-gamma*beta)*np.dot(ainv, b)
    cc = np.dot(c, ainv)
    dc = d + gamma*np.linalg.multi_dot([c, ainv, b])
    return ac, bc, cc, dc

def moebius_arg(z, par = (1, 0, 0, 1)):
    alpha, beta, gamma, delta = par
    nz = np.shape(z)[0]
    s = np.empty(nz, dtype=complex)
    for idx in np.arange(nz):
        s[idx]= (alpha*z[idx]+beta)/(gamma*z[idx]+delta)
    return s

def moebius_inv(sys, par = (1, 0, 0, 1)):
    """ Calculates the bilinear transformation D->C for ss-system sys """
    a, b, c, d = sys
    alpha, beta, gamma, delta = par
    n =  np.shape(a)[0]
    ainv = np.linalg.inv(delta*np.eye(n)+gamma*a)
    ac = np.dot((alpha*a +beta*np.eye(n)), ainv)
    bc = np.dot(ainv, b)
    cc = -(gamma*beta-alpha*delta)*np.dot(c, ainv)
    dc = d - gamma*np.linalg.multi_dot([c, ainv, b])
    return ac, bc, cc, dc
    alpha, beta, gamma, delta = par

def moebius_arg_inv(s, par = (1, 0, 0, 1)):
    alpha, beta, gamma, delta = par
    nz = np.shape(s)[0]
    z = np.empty(nz, dtype=complex)
    for idx in np.arange(nz):
        z[idx]= (beta-delta*s[idx])/(gamma*s[idx]-alpha)
    return z

def uq_cond(z, q):
    m = 1
    nw = np.size(z, 0)
    u = np.empty([m*q, nw*m], dtype='complex')
    for widx in range(nw):
        u[:m, widx*m:(widx+1)*m] = np.eye(m)
        zx = z[widx]
        for qidx in range(q)[1:]:
            u[qidx*m:(qidx+1)*m, widx*m:(widx+1)*m] = zx*np.eye(m)
            zx *= z[widx]
    return np.linalg.cond(u)

if __name__ == "__main__":

# Below is unit test code
    def unit_test_markov_kung():
        nmpset = [(4, 1, 1), (1, 1, 1), (4, 2, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            a = np.random.randn(n, n)
            b = np.random.randn(n, m)
            c = np.random.randn(p, n)
            mp = markov((a, b, c), 2*n )
            ae ,be, ce = kung_realization(mp, n)
            me = markov((ae, be, ce), 2*n )
            err = linalg.norm(me-mp)/linalg.norm(mp)
            if err > 1e-8:
                print('Unit test "markov_kung" failed', err, n, m, p)
                return False
        print('Unit test "markov_kung" passed')
        return True

    def unit_test_ls_estim_cd():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            z = np.exp(1j*2 * np.pi * fset)
            fd = fresp(z, A, B, C, D)
            Ce, De = ls_estim_cd(fd, z, A, B)
            fde = fresp(z, A, B, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "ls_estim_cd 1" failed')
                return False
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.zeros((p, m))
            fset = np.arange(0, N, dtype='float')/N
            z = np.exp(1j*2 * np.pi * fset)
            fd = fresp(z, A, B, C, D)
            Ce, De = ls_estim_cd(fd, z, A, B, estimd=False)
            fde = fresp(z, A, B, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "ls_estim_cd 2" failed')
                return False
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n) + 1j*np.random.randn(n, n)
            B = np.random.randn(n, m) + 1j*np.random.randn(n, m)
            C = np.random.randn(p, n) + 1j*np.random.randn(p, n)
            D = np.random.randn(p, m) + 1j*np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            z = np.exp(1j*2 * np.pi * fset)
            fd = fresp(z, A, B, C, D)
            Ce, De = ls_estim_cd(fd, z, A, B, dtype='complex')
            fde = fresp(z, A, B, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "ls_estim_cd 3" failed')
                return False
        print('Unit test "ls_estim_cd" passed')
        return True

    def unit_test_ls_estim_bd():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            z = np.exp(1j*2 * np.pi * fset)
            fd = fresp(z, A, B, C, D)
            Be, De = ls_estim_bd(fd, z, A, C)
            fde = fresp(z, A, Be, C, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "ls_estim_bd" failed')
                return False
        print('Unit test "ls_estim_bd" passed')
        return True

    def unit_test_transp_ffdata():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            z = np.exp(1j*2 * np.pi * fset)
            fd = fresp(z, A, B, C, D)
            fde = transp_ffdata(transp_ffdata(fd))
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "transp_fdata" failed')
                return False
        print('Unit test "transp_fdata" passed')
        return True

    def unit_test_fresp():
        nw = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            T = np.random.randn(n, n)
            s = linalg.svd(T, compute_uv=False)
            if s[n-1] < 1e-6:
                T = T + 1e-6*np.eye(n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            dd = np.zeros(n,dtype=complex)
            fset = np.arange(0, nw, dtype='float')/nw
            z = np.exp(1j*2 * np.pi * fset)
            adiag = np.random.randn(n)
            A = np.diag(adiag)
            fd = np.empty([nw, p, m], dtype='complex')
            for fidx in range(nw):
                emad = z[fidx] * np.ones(n)
                for didx in range(n):
                    dd[didx] = 1/(emad[didx]-adiag[didx])
                fd[fidx, :, :] = np.dot(np.dot(C, np.diag(dd)), B) + D
            Tinv = linalg.inv(T)
            Ae = np.dot(T, np.dot(A, Tinv))
            Be = np.dot(T, B)
            Ce = np.dot(C, Tinv)
            De = D
            fde = fresp_slow(z, Ae, Be, Ce, De)
            fdef = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            errf = linalg.norm(fd-fdef)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fresp_slow" failed')
                return False
            if errf > 1e-8:
                print('Unit test "fresp" failed', errf)
                return False
        print('Unit test "fresp" "fresp_slow" passed')
        return True

    def unit_test_fresp_def():
        N = 100
        nmpset = [(2, 4, 12), (2, 3, 6)]
        for (n, m, p) in nmpset:
            A = np.array([[0, 0], [1, 0]])
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N+1.0e-5
            z = np.exp(1j*2*np.pi*fset)
            frsp = fresp_slow(z, A, B, C, D)
            frspf = fresp(z, A, B, C, D)
            err = linalg.norm(frspf-frsp)/linalg.norm(frsp)
            if err > 1e-8:
                print('Unit test "fresp_def" failed for a defective matrix', err)
                return False
        print('Unit test "fresp_def" passed')
        return True


    def unit_test_fdestim_bd():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Be, De, resid = fdestim_bd(zn, Y, U, A, C)
            fde = fresp(z, A, Be, C, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_bd" failed')
                return False
        print('Unit test "fdestim_bd" passed')
        return True

    def unit_test_fdestim_bd_w():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            W = np.empty((N*m,p,p));
            for nidx in range(N*m):
                W[nidx,:,:] = np.random.randn(p,p)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Be, De, resid = fdestim_bd(zn, Y, U, A, C, w=W)
            fde = fresp(z, A, Be, C, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_bd_w" failed')
                return False
            B = np.random.randn(n, m)+1j*np.random.randn(n, m)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Be, De, resid = fdestim_bd(zn, Y, U, A, C, dtype=complex, w=W)
            fde = fresp(z, A, Be, C, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_bd_w" failed')
                return False
        print('Unit test "fdestim_bd_w" passed')
        return True

    def unit_test_fdestim_bd_no_d():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.zeros((p, m))
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Be, De, resid = fdestim_bd(zn, Y, U, A, C, estimd=False)
            fde = fresp(z, A, Be, C, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_bd_no_d" failed')
                return False
        print('Unit test "fdestim_bd_no_d" passed')
        return True


    def unit_test_fdestim_bd_cmplx():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            B = B+1j*np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            D = D+ 1j*np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Be, De = fdestim_bd(zn, Y, U, A, C, dtype='complex')
            fde = fresp(z, A, Be, C, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_bd cmplx" failed')
                return False
        print('Unit test "fdestim_bd complex" passed')
        return True

    def unit_test_fdestim_cd():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Ce, De, resid = fdestim_cd(zn, Y, U, A, B)
            fde = fresp(z, A, B, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_cd" failed')
                return False
        print('Unit test "fdestim_cd" passed')
        return True

    def unit_test_fdestim_cd_no_d():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.zeros((p,m))
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Ce, De, resid = fdestim_cd(zn, Y, U, A, B, estimd=False)
            fde = fresp(z, A, B, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_cd_no_d" failed')
                return False
        print('Unit test "fdestim_cd_no_d" passed')
        return True



    def unit_test_fdestim_cd_w():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            W = np.empty((N*m,p,p));
            for nidx in range(N*m):
                W[nidx,:,:] = np.random.randn(p,p)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Ce, De, resid = fdestim_cd(zn, Y, U, A, B, w=W)
            fde = fresp(z, A, B, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_cd_w" failed')
                return False
            C = np.random.randn(p, n)+1j*np.random.randn(p, n)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            zn = np.exp(1j*wn)
            Ce, De, resid = fdestim_cd(zn, Y, U, A, B, dtype='complex', w=W)
            fde = fresp(z, A, B, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "fdestim_cd_w" failed')
                return False

        print('Unit test "fdestim_cd_w" passed')
        return True


    def unit_test_ltifr_slow():
        N = 100
        nw = N
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12), (2, 3, 12)]
        for (n, m, p) in nmpset:
            dd = np.zeros(n,dtype=complex)
            T = np.random.randn(n, n)
            s = linalg.svd(T, compute_uv=False)
            if s[n-1] < 1e-6:
                T = T + 1e-6*np.eye(n)
            B = np.random.randn(n, m)
            fset = np.arange(0, N, dtype='float')/N
            z = np.exp(1j*2*np.pi*fset)
            adiag = np.random.randn(n)
            A = np.diag(adiag)
            Tinv = linalg.inv(T)
            fkern1 = np.empty([nw, n, m], dtype='complex')
            for fidx in range(N):
                emad = z[fidx]*np.ones(n)
                for didx in range(n):
                    dd[didx] = 1/(emad[didx]-adiag[didx])
                fkern1[fidx, :, :] = np.dot(T, np.dot( np.diag(dd), B))
#
#
#                fkern1[fidx, :, :] = np.dot(T, np.dot(np.diag(
#                    map(lambda x: 1/x, emad-adiag)), B))
            Ae = np.dot(T, np.dot(A, Tinv))
            Be = np.dot(T, B)
            fkern = ltifr_slow(Ae, Be, z)
            fkernf = ltifr(Ae, Be, z)
            err = linalg.norm(fkern1-fkern)/linalg.norm(fkern1)
            errf = linalg.norm(fkern1-fkernf)/linalg.norm(fkern1)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "ltifr_slow" failed', err)
                return False
            if errf > 1e-8:
                print('Unit test "ltifr" failed', errf)
                return False
        print('Unit test "ltifr_slow" passed')
        print('Unit test "ltifr" passed')
        return True

    def unit_test_ltifr_def():
        N = 100
        nmpset = [(2, 4, 12), (2, 3, 6)]
        for (n, m, p) in nmpset:
            A = np.array([[0, 0], [1, 0]])
            B = np.random.randn(n, m)
            fset = np.arange(0, N, dtype='float')/N+1.0e-5
            z = np.exp(1j*2*np.pi*fset)
            fkern = ltifr_slow(A, B, z)
            fkernf = ltifr(A, B, z)
            err = linalg.norm(fkernf-fkern)/linalg.norm(fkern)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "ltifr_def" failed for defective a matrix', err)
                return False
        print('Unit test "ltifr_def" passed')
        return True

    def unit_test_bilinear():
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        T = 2
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            a, b, c, d = bilinear_c2d((A, B, C, D), T)
            Ae, Be, Ce, De = bilinear_d2c((a, b, c, d), T)
            err = (linalg.norm(A-Ae)/linalg.norm(A) +
                linalg.norm(B-Be)/linalg.norm(B) +
                linalg.norm(C-Ce)/linalg.norm(C) +
                linalg.norm(D-De)/linalg.norm(D))
            if err > 1e-8:
                print('Unit test "bilinear 1" failed err=', err)
                return False
            N = 100
            fset = np.arange(0, N, dtype='float')/N
            s = 1j*2*np.pi*fset
            z = np.exp(1j*cf2df(2*np.pi*fset,T))
            Hc = fresp(s, A, B, C, D)
            Hd = fresp(z, a, b, c, d)
            err = linalg.norm(Hc-Hd)/linalg.norm(Hc)
            if err > 1e-8:
                print('Unit test "bilinear 2" failed err=', err)
                return False
        print('Unit test "bilinear 1 and 2" passed')
        return True

    def unit_test_fconv():
        N = 100
        T = 100
        fset = np.arange(-(N-1), N-1, dtype='float')
        wc = fset
        wd = cf2df(wc,T)
        w = df2cf(wd,T)
        err = linalg.norm(wc-w)/linalg.norm(wc)
        if err > 1e-8:
            print('Unit test "fconv" failed')
            return False
        print('Unit test "fconv" passed')
        return True


    def unit_test_ffsid():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            Ae, Be, Ce, De, s =  ffsid(w, fd, n, n+1, dtype='float', estimd=True)
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "ffsid" failed')
                return False
        print('Unit test "ffsid" passed')
        return True
    def unit_test_ffsid_no_d():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
          #  D = np.random.randn(p, m)
            D = np.zeros((p,m))
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            Ae, Be, Ce, De, s =  ffsid(w, fd, n, n+1, dtype='float', estimd=False)
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "ffsid_no_d" failed')
                return False
        print('Unit test "ffsid_no_d" passed')
        return True
    def unit_test_ffsid_complex():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            A = A + 1j*np.random.randn(n, n)
            B = np.random.randn(n, m)
            B = B+1j*np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            D = D+ 1j*np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            Ae, Be, Ce, De, s =  ffsid(w, fd, n, n+1, dtype='complex', estimd=True)
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "ffsid complex" failed')
                return False
        print('Unit test "ffsid complex" passed')
        return True
    def unit_test_fdsid():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            lam = linalg.eig(A)[0]
            rho = np.max( np.abs(lam))
            ## Here we create a random stable DT system
            A = A/rho/1.01
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            u = np.random.randn(N, m)
            y = lsim((A, B, C, D), u, dtype='float')
            yf = np.fft.fft(y,axis=0)
            uf = np.fft.fft(u,axis=0)
            fddata = (w, yf, uf)
            Ae, Be, Ce, De, xt, s =  fdsid(fddata, n, 2*n,
                       estTrans=True, dtype='float')
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "fdsid" failed')
                return False
            W = np.random.randn(N,p,p)
#            for idx in range(N):
#                W[idx,:,:] = np.eye(p)
            Ae, Be, Ce, De, xt, s =  fdsid(fddata, n, 2*n,
                                                estTrans=True, dtype='float', W=W)
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "fdsid with weighting" failed')
                return False
        print('Unit test "fdsid" passed')
        return True

    def unit_test_fdsid_noTrans():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            U, Y, wn = ffdata2fddata(fd, w)
            fddata = (wn, Y, U)
            Ae, Be, Ce, De, xt, s =  fdsid(fddata, n, 2*n,
                       estTrans=False, dtype='float')
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "fdsid_noTrans" failed')
                return False
            W = np.random.randn(N*m,p,p)
#            for idx in range(N):
#                W[idx,:,:] = np.eye(p)
            Ae, Be, Ce, De, xt, s =  fdsid(fddata, n, 2*n,
                                                estTrans=False, dtype='float', W=W)
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "fdsid_notrans with weighting" failed')
                return False
        print('Unit test "fdsid_notrans" passed')
        return True



    def unit_test_fdsid_no_d():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            lam = linalg.eig(A)[0]
            rho = np.max( np.abs(lam))
            ## Here we create a random stable DT system
            A = A/rho/1.01
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.zeros((p, m))
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            u = np.random.randn(N, m)
            y = lsim((A, B, C, D), u, dtype='float')
            yf = np.fft.fft(y,axis=0)
            uf = np.fft.fft(u,axis=0)
            fddata = (w, yf, uf)
            Ae, Be, Ce, De, xt, s =  fdsid(fddata, n, 2*n,
                       estTrans=True, dtype='float', estimd=False)
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "fdsid_no_d" failed')
                return False
        print('Unit test "fdsid_no_d" passed')
        return True

    def unit_test_fdsid_cmplx():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            lam = linalg.eig(A)[0]
            rho = np.max( np.abs(lam))
            ## Here we create a random stable DT system
            A = A/rho/1.01
            B = np.random.randn(n, m)
            B = B+1j*np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            D = D+ 1j*np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            w = 2*np.pi*fset
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            u = np.random.randn(N, m)
            y = lsim((A, B, C, D), u, dtype='complex')
            yf = np.fft.fft(y,axis=0)
            uf = np.fft.fft(u,axis=0)
            fddata = (w, yf, uf)
            Ae, Be, Ce, De, xt, s =  fdsid(fddata, n, 2*n,
                                                estTrans=True, dtype='complex')
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "fdsid_cmplx" failed')
                return False
            W = np.random.randn(N,p,p)
#            for idx in range(N):
#                W[idx,:,:] = np.eye(p)
            Ae, Be, Ce, De, xt, s =  fdsid(fddata, n, 2*n,
                                                estTrans=True, dtype='complex', W=W)
            fde = fresp(z, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            if err > 1e-8:
                print('Unit test "fdsid_complex with weighting" failed')
                return False
        print('Unit test "fdsid_complex" passed')
        return True

    def unit_test_lrm():
        N = 400
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            lam = linalg.eig(A)[0]
            rho = np.max( np.abs(lam))
            ## Here we create a random stable DT system
            A = A/rho/1.01
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            fset = np.arange(0, N, dtype='float')/N
            z = np.exp(1j*2*np.pi*fset)
            fd = fresp(z, A, B, C, D)
            u = np.random.randn(N, m)
            y = lsim((A, B, C, D), u, dtype='float')
            ff = lrm(u,y)
            err = linalg.norm(fd-ff)/linalg.norm(fd)
            if err > 1e-3:
                print('Unit test "lrm 1" failed')
                return False
            ff = lrm(u,y, lam=1e-7)
            err = linalg.norm(fd-ff)/linalg.norm(fd)
            if err > 1e-3:
                print('Unit test "lrm 2" failed')
                return False
            DFTind = np.arange(10,30,2)
            ff = lrm(u,y, DFTindices = DFTind, lam=1e-7)
            err = linalg.norm(fd[DFTind,: , :]-ff)/linalg.norm(fd[DFTind,: , :])
            if err > 1e-3:
                print('Unit test "lrm 3" failed')
                return False
        print('Unit test "lrm" passed')
        return True

    def unit_test_moebius():
        N = 100
        nmpset = [(4, 1, 1), (1, 1, 1), (2, 4, 12)]
        sset = np.random.randn(N)+1j*np.random.randn(N)
        par = np.random.randn(4)+1j*np.random.randn(4)
        zset = moebius_arg_inv(sset, par)
        sset1 = moebius_arg(zset, par)
        if linalg.norm(sset-sset1)/linalg.norm(sset)>1e-8:
            print('Unit test "moebius 0" failed')
            return False
        for (n, m, p) in nmpset:
            A = np.random.randn(n, n)
            B = np.random.randn(n, m)
            C = np.random.randn(p, n)
            D = np.random.randn(p, m)
            zset = np.random.randn(N)+1j*np.random.randn(N)
            par = np.random.randn(4)+1j*np.random.randn(4)
            sset = moebius_arg(zset, par)
            fd = fresp(sset, A, B, C, D)
            Ae, Be, Ce, De = moebius((A,B,C,D), par)
            fde = fresp(zset, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "moebius 1" failed')
                print('err= ',err)
                return False
            sset = np.random.randn(N)+1j*np.random.randn(N)
            zset = moebius_arg_inv(sset, par)
            fd = fresp(zset, A, B, C, D)
            Ae, Be, Ce, De = moebius_inv((A,B,C,D), par)
            fde = fresp(sset, Ae, Be, Ce, De)
            err = linalg.norm(fd-fde)/linalg.norm(fd)
            # print('|| H-He ||/||H|| = ', err)
            if err > 1e-8:
                print('Unit test "moebius 2" failed')
                print('err= ',err)
                return False
            Ae, Be, Ce, De = moebius((A,B,C,D), par)
            a, b, c, d = moebius_inv((Ae,Be,Ce,De), par)
            sys = (a,b,c,d)
            sys0 = (A,B,C,D)
            err = 0
            for idx in np.arange(4):
                err += linalg.norm(sys[idx]-sys0[idx])/linalg.norm(sys0[idx])
            if err > 1e-8:
                print('Unit test "moebius 3" failed')
                print('err= ',err)
                return False

        print('Unit test "moebius" passed')
        return True



# Run the unit tests

    fsid()
    if (unit_test_fdestim_bd()
        and unit_test_fdestim_bd_no_d()
        and unit_test_fdestim_bd_w()
        and unit_test_fdestim_cd()
        and unit_test_fdestim_cd_no_d()
        and unit_test_fdestim_cd_w()
        and unit_test_fresp()
        and unit_test_fresp_def()
        and unit_test_ls_estim_bd()
        and unit_test_ls_estim_cd()
        and unit_test_transp_ffdata()
        and unit_test_fdsid()
        and unit_test_fdsid_noTrans()
        and unit_test_fdsid_no_d()
        and unit_test_fdsid_cmplx()
        and unit_test_ltifr_slow()
        and unit_test_ltifr_def()
        and unit_test_ffsid()
        and unit_test_ffsid_no_d()
        and unit_test_ffsid_complex()
        and unit_test_bilinear()
        and unit_test_fconv()
        and unit_test_lrm()
        and unit_test_moebius()
        and unit_test_markov_kung()):
        print("All unit tests passed")


"""Centralized default constants and types used across the package."""

from typing import Literal

import jax
import optimistix as optx


DeviceName = Literal["cpu", "gpu", "tpu"]
DeviceLike = DeviceName | jax.Device | None


# When changing these constants, also update the corresponding docstrings!
PRINT_EVERY = 1
SEED = 42
SOLVER = optx.LevenbergMarquardt(rtol=1e-3, atol=1e-6)



"""Data structures in time and frequency domains, including metadata."""

from dataclasses import dataclass

import jax.numpy as jnp
import numpy as np

from . import _best_linear_approximation


@dataclass(frozen=True)
class TimeData:
    """Normalized time-domain data container.

    Attributes
    ----------
    u : jnp.ndarray, shape (N, nu, R)
        Normalized input signals, averaged over periods.
    y : jnp.ndarray, shape (N, ny, R)
        Normalized output signals, averaged over periods.
    t : np.ndarray, shape (N,)
        Time vector of a single period.
    ts : float
        Sampling time in seconds.

    """

    u: jnp.ndarray
    y: jnp.ndarray
    t: np.ndarray
    ts: float


@dataclass(frozen=True)
class NonparametricBLA:
    """Nonparametric Best Linear Approximation (BLA) and distortion levels.

    Attributes
    ----------
    G : np.ndarray
        Nonparametric BLA estimate, shape (F, ny, nu).
    var_noise : np.ndarray, shape (F, ny, nu), optional
        Estimated noise variance per period. Is `None` if `P == 1`.
    var_tot : np.ndarray, shape (F, ny, nu), optional
        Estimated total variance per experiment (`R // nu`). Is `None`
        if `R // nu == 1`.

    """

    G: np.ndarray
    var_noise: np.ndarray | None
    var_tot: np.ndarray | None


@dataclass(frozen=True)
class FrequencyData:
    """Normalized frequency-domain data container.

    Attributes
    ----------
    G_bla : `NonparametricBLA`, optional
        Nonparametric BLA estimate with variance estimates. Is `None` if
        insufficient realizations are available to compute the frequency
        response matrix (i.e., `R < nu`).
    U : jnp.ndarray, shape (N//2 + 1, nu, R)
        Normalized input DFT, averaged over periods.
    Y : jnp.ndarray, shape (N//2 + 1, ny, R)
        Normalized output DFT, averaged over periods.
    Y_var_noise : jnp.ndarray, shape (F, ny), optional
        Estimated output noise variance per period. Is `None` if `P == 1`.
    f : np.ndarray, shape (N//2 + 1,)
        Complete frequency vector.
    f_idx : np.ndarray, shape (F,)
        Excited frequency indices.
    fs : float
        Sampling frequency in Hz.

    """

    G_bla: NonparametricBLA | None
    U: jnp.ndarray
    Y: jnp.ndarray
    Y_var_noise: jnp.ndarray | None
    f: np.ndarray
    f_idx: np.ndarray
    fs: float


@dataclass(frozen=True)
class Normalizer:
    """Normalization statistics for input/output signals.

    Attributes
    ----------
    u_mean : np.ndarray, shape (nu,)
        Input means.
    u_std : np.ndarray, shape (nu,)
        Input standard deviations.
    y_mean : np.ndarray, shape (ny,)
        Output means.
    y_std : np.ndarray, shape (ny,)
        Output standard deviations.

    """

    u_mean: np.ndarray
    u_std: np.ndarray
    y_mean: np.ndarray
    y_std: np.ndarray


@dataclass(frozen=True)
class InputOutputData:
    """Combined time and frequency domain data.

    Attributes
    ----------
    time : `TimeData`
    freq : `FrequencyData`
    norm : `Normalizer`

    """

    time: TimeData
    freq: FrequencyData
    norm: Normalizer


def create_data_object(
    u: np.ndarray,
    y: np.ndarray,
    f_idx: np.ndarray,
    fs: float
) -> InputOutputData:
    """Create InputOutputData object from time-domain signals and frequency metadata.

    Parameters
    ----------
    u : np.ndarray, shape (N, nu, R, P)
        Input time series, a (random-phase) (multi)sine with 4 dimensions:
        - N: Number of samples per period;
        - nu: Number of input channels;
        - R: Number of independent phase realizations. Each realization must
             have the same frequency content and amplitude characteristics;
        - P: Number of periods (copies of the same realization).
        It is fine if only a single realization and/or period is provided,
        but is is important to shape the data in the required 4D format.
    y : np.ndarray, shape (N, ny, R, P)
        Output time series. Similar structure as `u`, with ny as the number of
        output channels. All periods must be in steady-state,
        i.e., there should be almost no transient effects.
    f_idx : np.ndarray, shape (F,)
        Indices of excited frequencies, where `F ≤ N//2 + 1`.
    fs : float
        Sampling frequency in Hz.

    Returns
    -------
    InputOutputData
        Processed (meta)data in time and frequency domains.

    """
    # Validate dimensions
    if u.ndim != 4:
        raise ValueError("`u` must have 4 dimensions: (N, nu, R, P).")
    if y.ndim != 4:
        raise ValueError("`y` must have 4 dimensions: (N, ny, R, P).")
    if u.shape[0] != y.shape[0]:
        raise ValueError("`u` and `y` must have same number of time samples.")
    if u.shape[2] != y.shape[2]:
        raise ValueError("`u` and `y` must have same number of realizations.")
    if u.shape[3] != y.shape[3]:
        raise ValueError("`u` and `y` must have same number of periods.")

    ts = 1 / fs
    N, nu, R, P = u.shape
    t = np.arange(N) * ts

    # Normalize data (zero mean, unit variance)
    u_mean = u.mean(axis=(0, 2, 3), keepdims=True)
    y_mean = y.mean(axis=(0, 2, 3), keepdims=True)
    u_std = u.std(axis=(0, 2, 3), keepdims=True)
    y_std = y.std(axis=(0, 2, 3), keepdims=True)

    u = (u - u_mean) / u_std
    y = (y - y_mean) / y_std

    # Compute DFTs
    U = np.fft.rfft(u, axis=0)
    Y = np.fft.rfft(y, axis=0)
    f = np.arange(N//2 + 1) * fs / N

    # Compute output noise variance (will be used in loss functions)
    Y_avg = Y.mean(axis=3)
    if P > 1:
        sqr_error = np.abs(Y - Y_avg[..., None])**2  # shape (F, ny, R, P)
        tot_sqr_error = sqr_error.sum(axis=(2, 3))  # shape (F, ny)
        Y_var_noise = tot_sqr_error / (R * (P - 1))  # shape (F, ny)
        Y_var_noise = jnp.asarray(Y_var_noise)
    else:
        Y_var_noise = None

    # Compute nonparametric BLA
    if R >= nu:
        G_bla = _best_linear_approximation.nonparametric_bla(U[f_idx], Y[f_idx])
    else:
        print(
            "Warning: Insufficient realizations (R < nu) to compute the nonparametric "
            "BLA. Identification can proceed, but the initial linear model may be "
            "suboptimal."
        )
        G_bla = None
        
    # We proceed with data that is averaged over periods
    u_avg, y_avg = u.mean(axis=3), y.mean(axis=3)
    U_avg = U.mean(axis=3)

    # Finally, we convert the input-output data to Jax arrays
    u_avg, y_avg = jnp.asarray(u_avg), jnp.asarray(y_avg)
    U_avg, Y_avg = jnp.asarray(U_avg), jnp.asarray(Y_avg)

    return InputOutputData(
        TimeData(u_avg, y_avg, t, ts),
        FrequencyData(G_bla, U_avg, Y_avg, Y_var_noise, f, f_idx, fs),
        Normalizer(u_mean.flatten(), u_std.flatten(),
                   y_mean.flatten(), y_std.flatten())
    )



"""Nonparametric BLA, parametric subspace identification, and optimizer."""
import equinox as eqx
import jax.numpy as jnp
import numpy as np
import optimistix as optx

from . import _misc
from ._config import PRINT_EVERY, SOLVER, DeviceLike
from ._data_manager import FrequencyData, InputOutputData, NonparametricBLA
from ._model_structures import ModelBLA
from ._solve import SolveResult, solve
from .dep import fsid


MAX_ITER = 1000  # when changing, also update corresponding docstring!


### Public functions ###


def nonparametric_bla(U: np.ndarray, Y: np.ndarray) -> NonparametricBLA:
    """Compute nonparametric BLA and variance estimates from input-output data.

    Parameters
    ----------
    U : np.ndarray, shape (F, nu, R, P)
        DFT input spectrum at the excited frequencies across realizations and periods.
    Y : np.ndarray, shape (F, ny, R, P)
        DFT output spectrum at the excited frequencies across realizations and periods.

    Returns
    -------
    `NonparametricBLA`
        Nonparametric BLA estimate with frequency response and variance estimates.
        
    Raises
    ------
    ValueError
        If the number of realizations R is less than the number of inputs nu.

    """
    nu, R = U.shape[1:3]
    if R < nu:
        raise ValueError(
            "For multi-input systems, the number of realizations (R) must be "
            "at least equal to the number of inputs (nu) to compute the "
            "frequency response matrix."
        )
    
    G = _compute_frequency_response(U, Y)  # shape (F, ny, nu, M, P)
    M, P = G.shape[3:5]

    # Compute noise variance
    G_P = G.mean(axis=4)  # shape (F, ny, nu, M)
    if P > 1:
        sqr_error = np.abs(G - G_P[..., None]) ** 2  # shape (F, ny, nu, M, P)
        tot_sqr_error = sqr_error.sum(axis=(3, 4))  # shape (F, ny, nu)
        var_noise = tot_sqr_error / (M * (P - 1))  # shape (F, ny, nu)
        var_noise = jnp.asarray(var_noise)
    else:
        var_noise = None

    # Compute total variance
    G_bla = G_P.mean(axis=3)  # shape (F, ny, nu)
    if M > 1:
        sqr_error = np.abs(G_P - G_bla[..., None]) ** 2  # shape (F, ny, nu, M)
        tot_sqr_error = sqr_error.sum(axis=3)  # shape (F, ny, nu)
        var_tot = tot_sqr_error / (M - 1)  # shape (F, ny, nu)
        var_tot = jnp.asarray(var_tot)
    else:
        var_tot = None

    G_bla = jnp.asarray(G_bla)

    return NonparametricBLA(G_bla, var_noise, var_tot)


def subspace_id(
    data: InputOutputData,
    nx: int,
    nq: int | None = None,
    freq_weighting: bool = True,
    input_output_mode: bool = False,
    logging_enabled: bool = True
) -> ModelBLA:
    """Parametrize a state-space model using the frequency-domain subspace method.

    Parameters
    ----------
    data : `InputOutputData`
        Estimation data.
    nx : int
        State dimension of the system to be identified.
    nq : int | None, optional
        Subspace dimensioning parameter, must be greater than `nx`. Defaults to
        `nx + 1` if not provided.
    freq_weighting : bool
        Whether to use frequency weighting based on the inverse of the total variance
        on the nonparametric BLA. Defaults to `True`.
    input_output_mode : bool
        Whether to parametrize the state-space model directly from the input-output
        spectra instead of the nonparametric BLA. This mode is automatically activated 
        if no BLA estimate is available, even if `input_output_mode` is set to `False`. 
        Defaults to `False`.
    logging_enabled : bool
        Whether to print a summary of the identification results. Defaults to `True`.

    Returns
    -------
    `ModelBLA`
        Estimated state-space model in BLA form.

    Raises
    ------
    ValueError
        If `nq` is not greater than `nx`.

    """
    if logging_enabled:
        header = " Frequency-domain subspace identification "
        print(f"{header:=^72}")
    
    nq = nx + 1 if nq is None else nq
    if nq <= nx:
        raise ValueError(
            f"Subspace dimension nq={nq} must be greater than state dimension nx={nx}."
        )

    freq_data = data.freq
    input_output_mode, freq_weighting = _validate_inputs(
        input_output_mode, freq_weighting, freq_data.G_bla, logging_enabled
    )
    
    # Run the subspace identification
    A, B_u, C_y, D_yu = _subspace_id(
        freq_data, nx, nq, freq_weighting, input_output_mode
    )
    model = ModelBLA(A, B_u, C_y, D_yu, 1 / freq_data.fs, data.norm)
    
    if logging_enabled:
        x_bla = _misc.compute_steady_state_bla_state(model, data)
        _misc.evaluate_model_performance(model, data, x0=x_bla[0, :, :], offset=0)

    return model


def optimize(
    model: ModelBLA,
    data: InputOutputData,
    *,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser = SOLVER,
    freq_weighting: bool = True,
    input_output_mode: bool = False,
    max_iter: int = MAX_ITER,
    print_every: int = PRINT_EVERY,
    return_solve_details: bool = False,
    device: DeviceLike = None
) -> ModelBLA | tuple[ModelBLA, SolveResult]:
    """Refine the parameters of the BLA using frequency-response computations.

    Parameters
    ----------
    model : `ModelBLA`
        Initial BLA model to be optimized.
    data : `InputOutputData`
        Estimation data.
    solver : `optx.AbstractLeastSquaresSolver` or `optx.AbstractMinimiser`
        Any least-squares solver or general minimization solver from the
        Optimistix or Optax libraries. Defaults to 
        `optx.LevenbergMarquardt(rtol=1e-3, atol=1e-6)`.
    freq_weighting : bool
        Whether to use frequency weighting based on the inverse of the total variance
        on the nonparametric BLA. Defaults to `True`.
    input_output_mode : bool
        Whether to optimize the state-space model directly from the input-output
        spectra instead of the nonparametric BLA. This mode is automatically activated 
        if no BLA estimate is available, even if `input_output_mode` is set to `False`. 
        Defaults to `False`.
    max_iter : int
        Maximum number of optimization iterations. Defaults to `1000`.
    print_every : int
        Frequency of printing iteration information. If set to `0`, only a
        summary is printed. If set to `-1`, no printing is done. Defaults to `1`.
    return_solve_details : bool
        Whether to return detailed information about the optimization process. This is
        useful for e.g. plotting convergence curves. Defaults to `False`.
    device : `DeviceLike`, optional
        Device on which to perform the computations. Can be either a device
        name (`"cpu"`, `"gpu"`, or `"tpu"`) or a specific JAX device. If not
        provided, the default JAX device is used.
        
    Returns
    -------
    `ModelBLA`
        BLA model with optimized parameters.
    `SolveResult`, optional
        More details about the optimization process, only returned if
        `return_solve_details` is `True`.

    """
    logging_enabled = print_every != -1
    
    if logging_enabled:
        header = " BLA optimization "
        print(f"{header:=^72}")
    
    input_output_mode, freq_weighting = _validate_inputs(
        input_output_mode, freq_weighting, data.freq.G_bla, logging_enabled
    )
    
    # Run the optimization
    model, solve_result = _optimize(
        model, data, input_output_mode, freq_weighting,
        logging_enabled, solver, max_iter, print_every, device
    )

    if logging_enabled:
        x_bla = _misc.compute_steady_state_bla_state(model, data)
        _misc.evaluate_model_performance(
            model, data, x0=x_bla[0, :, :], offset=0, solve_result=solve_result
        )

    if return_solve_details:
        return model, solve_result
    return model


### Internal helpers ###


def _compute_frequency_response(U: np.ndarray, Y: np.ndarray) -> np.ndarray:
    """Compute frequency response matrix G(k) = Y(k) * (U(k))^(-1).

    Parameters
    ----------
    U : np.ndarray, shape (F, nu, R, P)
        DFT input spectra at the excited frequencies across realizations and periods.
    Y : np.ndarray, shape (F, ny, R, P)
        DFT output spectra at the excited frequencies across realizations and periods.

    Returns
    -------
    G : np.ndarray, shape (F, ny, nu, M, P)
        Frequency Response Matrix:
        - F: number of frequency bins;
        - ny: number of outputs;
        - nu: number of inputs;
        - M: number of experiments (R // nu);
        - P: number of periods.

    """
    F, nu, R, P = U.shape
    ny = Y.shape[1]

    M = R // nu
    if M * nu != R:
        print(
            "Warning: Suboptimal number of realizations. Not all realizations "
            "are used to compute the frequency response matrix. Ideally, "
            "the number of realizations (R) should be an integer multiple "
            "of the number of inputs (nu)."
        )

    G = np.zeros((F, ny, nu, M, P), dtype=complex)

    for kf in range(F):
        for kr in range(M):
            for kp in range(P):
                start_idx = kr * nu
                end_idx = (kr + 1) * nu
                U_block = U[kf, :, start_idx:end_idx, kp]
                Y_block = Y[kf, :, start_idx:end_idx, kp]

                U_inv = np.linalg.solve(U_block, np.eye(nu))
                G[kf, :, :, kr, kp] = Y_block @ U_inv

    return G


def _subspace_id(
    freq_data: FrequencyData,
    nx: int,
    nq: int,
    freq_weighting: bool,
    input_output_mode: bool
) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Perform actual subspace identification with validated inputs."""
    freqs = freq_data.f[freq_data.f_idx]
    fs = freq_data.fs
    z = 2 * np.pi * freqs / fs
    
    if input_output_mode:
        F = len(freqs)
        _, ny, R = freq_data.Y.shape
        nu = freq_data.U.shape[1]
        
        Y = np.transpose(freq_data.Y[freq_data.f_idx], (0, 2, 1)).reshape(R * F, ny)
        U = np.transpose(freq_data.U[freq_data.f_idx], (0, 2, 1)).reshape(R * F, nu) 
        zj = np.repeat(np.exp(z * 1j), R)
        W = np.empty(0)  # no weighting in input-output mode
        
    else:
        G_bla = freq_data.G_bla
        F, ny, nu = G_bla.G.shape

        # Convert BLA to "input-output form" for FSID algorithm compatibility
        Y = np.transpose(G_bla.G, (0, 2, 1)).reshape(nu * F, ny)
        U = np.tile(np.eye(nu), (F, 1))
        zj = np.repeat(np.exp(z * 1j), nu)

        # Create weighting matrix (inverse of total variance)
        if freq_weighting:
            W_temp = 1 / G_bla.var_tot

            # The four lines below are to ensure compatibility with fsid.gfdsid
            W_temp = np.transpose(np.sqrt(W_temp), (0, 2, 1)).reshape(nu * F, ny)
            W = np.zeros((nu * F, ny, ny))
            for k in range(nu * F):
                np.fill_diagonal(W[k], W_temp[k])
        else:
            W = np.empty(0)

    # Ensure that zj, Y, and U are NumPy (i.e., non-JAX) arrays
    zj = np.asarray(zj, dtype=np.complex128)
    Y = np.asarray(Y, dtype=np.complex128)
    U = np.asarray(U, dtype=np.complex128)

    # Perform frequency-domain subspace identification
    fddata = (zj, Y, U)
    A, B_u, C_y, D_yu = fsid.gfdsid(fddata=fddata, n=nx, q=nq, estTrans=False, w=W)[:4]
    return A, B_u, C_y, D_yu


def _optimize(
    model: ModelBLA,
    data: InputOutputData,
    input_output_mode: bool,
    freq_weighting: bool,
    logging_enabled: bool,
    solver: optx.AbstractLeastSquaresSolver | optx.AbstractMinimiser,
    max_iter: int,
    print_every: int,
    device: DeviceLike,
) -> tuple[ModelBLA, SolveResult]:
    """Perform actual optimization with validated inputs."""
    freqs = data.freq.f[data.freq.f_idx]
    model = _normalize_states(model, data)
    theta0, theta_static = eqx.partition(model, eqx.is_inexact_array)

    if input_output_mode:
        U_nonpar = jnp.asarray(data.freq.U)[data.freq.f_idx]
        Y_nonpar = jnp.asarray(data.freq.Y)[data.freq.f_idx]
        args = (theta_static, U_nonpar, Y_nonpar, freqs)
        loss_fn = _loss_output_spectrum
   
    else:
        G_bla = data.freq.G_bla
        
        # Create weighting matrix (inverse of total variance)
        if freq_weighting:
            W = 1 / G_bla.var_tot
        else:
            W = jnp.ones_like(G_bla.G)

        args = (theta_static, jnp.asarray(G_bla.G), freqs, W)
        loss_fn = _loss_frequency_response

    # Optimize the model parameters
    if logging_enabled:
        print("Starting iterative optimization...")
    solve_result = solve(theta0, solver, args, loss_fn, max_iter, print_every, device)

    model = eqx.combine(solve_result.theta, theta_static)
    model = _normalize_states(model, data)
    return model, solve_result


def _loss_frequency_response(theta_dyn: ModelBLA, args: tuple) -> tuple:
    """Compute the weighted loss between the nonparametric and parametric BLA."""
    theta_static, G_nonpar, freqs, W = args

    theta = eqx.combine(theta_dyn, theta_static)

    G_par = theta._frequency_response(freqs)
    loss = jnp.sqrt(W / G_nonpar.size) * (G_par - G_nonpar)
    return _misc.real_valued(loss), (_misc.scalar_valued(loss),)


def _loss_output_spectrum(theta_dyn: ModelBLA, args: tuple) -> tuple:
    """Compute the loss between the nonparametric and parametric output spectra."""
    theta_static, U_nonpar, Y_nonpar, freqs = args

    theta = eqx.combine(theta_dyn, theta_static)

    Y_par = theta._frequency_response(freqs) @ U_nonpar
    loss = jnp.sqrt(1 / Y_nonpar.size) * (Y_par - Y_nonpar)
    return _misc.real_valued(loss), (_misc.scalar_valued(loss),)


def _normalize_states(model: ModelBLA, data: InputOutputData) -> ModelBLA:
    """Normalize BLA model states to have unit variance."""
    nx, nu = model.B_u.shape
    N = data.time.u.shape[0]

    G_xu = ModelBLA(  # parametric u->x frequency response; not the true BLA
        A=model.A, B_u=model.B_u, C_y=np.eye(nx), D_yu=np.zeros((nx, nu)), 
        ts=model.ts, norm=model.norm,
    )._frequency_response(data.freq.f)  # shape (N//2 + 1, nx, nu)

    X = G_xu @ data.freq.U  # shape (N//2 + 1, nx, R)
    x = np.fft.irfft(X, n=N, axis=0)  # shape (N, nx, R)
    x_std = np.std(x, axis=(0, 2))

    Tx = np.diag(x_std)
    Tx_inv = np.diag(1 / x_std)

    # Apply similarity transformation: x_norm = Tx_inv * x
    return ModelBLA(
        A=Tx_inv @ model.A @ Tx,
        B_u=Tx_inv @ model.B_u,
        C_y=model.C_y @ Tx,
        D_yu=model.D_yu,
        ts=model.ts,
        norm=model.norm
    )


def _validate_weighting(
    freq_weighting: bool,
    G_bla: NonparametricBLA | None,
    input_output_mode: bool,
    print_warning: bool,
) -> bool:
    """Validate whether frequency weighting can be applied.

    Weighting is disabled if:
      - `input_output_mode` is active, or
      - `freq_weighting` is requested but BLA total variance is unavailable.
    """
    if not freq_weighting:
        return False

    # Case 1: Incompatible with input-output mode
    if input_output_mode:
        if print_warning:
            print(
                "Warning: Frequency weighting based on BLA total variance requested, "
                "but input-output mode is active. Proceeding without weighting."
            )
        return False

    # Case 2: BLA variance not available
    if G_bla is None or G_bla.var_tot is None:
        if print_warning:
            print(
                "Warning: Frequency weighting based on BLA total variance requested, "
                "but such estimate is not available. Proceeding without weighting."
            )
        return False

    return True


def _validate_inputs(
    input_output_mode: bool,
    freq_weighting: bool,
    G_bla: NonparametricBLA | None,
    print_warning: bool
) -> tuple[bool, bool]:
    """Validate inputs for subspace identification."""
    # Switch to input-output mode if BLA data is missing
    if G_bla is None and not input_output_mode:
        if print_warning:
            print(
                "Warning: Nonparametric BLA estimate is not available. Proceeding "
                "with input-output mode for subspace identification."
            )
        input_output_mode = True
    
    # Validate frequency weighting settings
    freq_weighting = _validate_weighting(
        freq_weighting, G_bla, input_output_mode, print_warning
    )
    return input_output_mode, freq_weighting



"""Quick example on the Silverbox benchmark dataset using a neural network."""

import jax

import freq_statespace as fss


data = fss.load_and_preprocess_silverbox_data()  # 8192 x 6 samples

# Step 1: BLA estimation
nx = 2  # state dimension
bla = fss.lin.subspace_id(data, nx)
bla = fss.lin.optimize(bla, data)

# Step 2: Nonlinear optimization
neural_net = fss.static.NeuralNetwork(
    nw=1, nz=1, layers=1, neurons_per_layer=10, activation=jax.nn.relu
)
nllfr = fss.nonlin.connect(bla, neural_net)
nllfr = fss.nonlin.optimize(nllfr, data, device="cpu")

# NOTE: My CPU is faster here because the optimization problem is recurrent in nature
# and has very little work per step, so GPU overhead dominates. Larger models perform 
# enough computation per step for GPUs to become advantageous.


"""Quick example on the Silverbox benchmark dataset using inference and learning."""

import time

import nonlinear_benchmarks
from nonlinear_benchmarks.error_metrics import RMSE

import freq_statespace as fss


# ========== SYSTEM IDENTIFICATION ==========
data = fss.load_and_preprocess_silverbox_data()

start_time = time.time()

# Step 1: BLA estimation
nx = 2  # state dimension
bla = fss.lin.subspace_id(data, nx)
bla = fss.lin.optimize(bla, data)

# Step 2: Inference and learning
phi = fss.static.basis.Polynomial(nz=1, degree=3)

nllfr = fss.nonlin.inference_and_learning(bla, data, phi, nw=1)

# Step 3: Nonlinear optimization
nllfr = fss.nonlin.optimize(nllfr, data, device="cpu")

# NOTE: My CPU is faster here because the optimization problem is recurrent in nature
# and has very little work per step, so GPU overhead dominates. Larger models perform 
# enough computation per step for GPUs to become advantageous.

total_time = time.time() - start_time
print(f"\nTotal time for training: {total_time:.2f} seconds")


# ========== TEST PERFORMANCE ==========

test = nonlinear_benchmarks.Silverbox()[1]
test_multisine, test_arrow_full, test_arrow_no_extrapolation = test

# -- Multisine test --
y_model = nllfr.simulate(test_multisine.u)[0]
n = test_multisine.state_initialization_window_length
rmse_ms = RMSE(test_multisine.y[n:], y_model[n:].flatten())
print(f"\nRMSE on multisine test: {1000 * rmse_ms:.3e} mV")

# -- Arrowhead full --
y_model = nllfr.simulate(test_arrow_full.u)[0]
n = test_arrow_full.state_initialization_window_length
rmse_af = RMSE(test_arrow_full.y[n:], y_model[n:].flatten())
print(f"RMSE on arrowhead full test: {1000 * rmse_af:.3e} mV")

# -- Arrowhead no extrapolation --
y_model = nllfr.simulate(test_arrow_no_extrapolation.u)[0]
n = test_arrow_no_extrapolation.state_initialization_window_length
rmse_an = RMSE(test_arrow_no_extrapolation.y[n:], y_model[n:].flatten())
print(f"RMSE on arrowhead no extrapolation: {1000 * rmse_an:.3e} mV")
